
#include "common/extern.h"
#include "common/subtasksupp.h"
#include "config.h"

#include <stdlib.h>

/// "Load (char *fname)"

int ConfigClass::
Load ()
{
  int i;
  char readbuffer[BUFFERSIZE], *cfgdata;
  BPTR cfgfile;

  if ((cfgfile = Open (filename, MODE_READWRITE)))
    {
      while (FGets (cfgfile, readbuffer, BUFFERSIZE) != NULL)
        {
          if ((i = strlen (readbuffer)) > 0)
            {
              if (readbuffer[i - 1] == '\x0a')
                {
                  readbuffer[i - 1] = '\0';
                }

              cfgdata = (char *) AllocVecPooled (mempool, i + 2);

              if (cfgdata == NULL)      /* allocation error */
                {
                  Close (cfgfile);
                  return (CFG_READ_FAIL);       /* return an error */
                }

              strncpy (cfgdata, readbuffer, i);
              cfgdata[i] = '\0';

              if (add_list_item (cfgdata) == FALSE)     /* error allocating node? */
                {
                  Close (cfgfile);      /* close the file */
                  return (CFG_READ_FAIL);       /* return an error */
                }
            }
        }

      /* if empty list add a clear line */
      if (cfglist.lh_Head == (struct Node *) &cfglist.lh_Tail)
        {
          cfgdata = (char *) AllocVecPooled (mempool, 4);

              if (cfgdata == NULL)      /* allocation error */
                {
                  Close (cfgfile);
                  return (CFG_READ_FAIL);       /* return an error */
                }

              strncpy (cfgdata, " \0", 2);

              if (add_list_item (cfgdata) == FALSE)     /* error allocating node? */
                {
                  Close (cfgfile);      /* close the file */
                  return (CFG_READ_FAIL);       /* return an error */
                }
        }


      Close (cfgfile);
    }

  return (CFG_READ_SUCCESS);
}

///

/// "Save ()"

int ConfigClass::
Save ()
{
  struct Node *worknode = cfglist.lh_Head;
  BPTR cfgfile = Open (filename, MODE_NEWFILE);
  char *tmp;

  if (cfgfile != NULL)
    {
      while (worknode->ln_Succ != NULL)
        {
          if ((tmp = worknode->ln_Name))
            {
              //strcat (worknode->ln_Name, "\n");
              while (*tmp != '\0')
                  tmp++;

              *tmp++ = '\n';
              *tmp   = '\0';

              if (FPuts (cfgfile, worknode->ln_Name) == -1)
                {
                  Close (cfgfile);
                  return (CFG_WRITE_FAIL);
                }
            }

          worknode = worknode->ln_Succ;
        }

      Close (cfgfile);
      return (CFG_WRITE_SUCCESS);
    }

  return (CFG_WRITE_FAIL);
}

///

/// "Constructor & destructor"
ConfigClass::ConfigClass(APTR pool, char *fname)
{
  mempool = pool;
  filename = fname;
	NewList (&cfglist);
}


ConfigClass::~ConfigClass()
{
  struct Node *worknode;

  while ((worknode = RemHead (&cfglist)) != NULL)
    {
      FreeVecPooled (mempool, worknode->ln_Name);
      FreeSPooled (mempool, worknode, sizeof (struct Node));
    }
}

///

/*

   Handling funcs

 */

/// "GetString (char *item, char *def)"
char *ConfigClass::
GetString (char *item, char *def)
{
  struct Node *worknode = cfglist.lh_Head;
  int ilen;
  char *line;


  /* Now scan the file looking the the desired section and item */
  while (worknode->ln_Succ != NULL)
    {
      if ((line = worknode->ln_Name))
        {
          // skip white spaces
          while (*line == ' ' || *line == '\t')
            ++line;

          if (*line != '\0')    /* (strlen (worknode->ln_Name) > 0) */
            {
              if ((strncmp (line, item, ilen = strlen (item))) == 0)
                {
                  line += ilen;

                  // now skip white spaces & search '='
                  while (*line == ' ' || *line == '\t')
                    ++line;

                  if (*line++ == '=')
                    {
                      while (*line == ' ' || *line == '\t')
                        ++line;

                      if (*line != '\0')        /* (strlen (tmp) > 0) */
                        return (line);
                    }
                }
            }
        }

      worknode = worknode->ln_Succ;
    }

  /* we've reached the end of the file without finding the requested item in */
  /* the requested section, so we'll return the default value instead. */

  return (def);
}

///

/// "SetString (char *item, char *def)"

BOOL ConfigClass::
SetString (char *item, char *def)
{
  char *tmpstring = NULL;
  struct Node *worknode = cfglist.lh_Head;
  int ilen;
  char *line;

  /* Now scan the file looking the desired section and item */
  while (worknode->ln_Succ != NULL)
    {
      if ((line = worknode->ln_Name))
        {
          while (*line == ' ' || *line == '\t')
            ++line;

          if (*line != '\0')    /* (strlen (worknode->ln_Name) > 0) */
            {
              if ((strncmp (line, item, ilen = strlen (item)) == 0))
                {
                  line += ilen;

                  while (*line == ' ' || *line == '\t')
                    ++line;

                  if (*line++ == '=')
                    {

                      while (*line == ' ' || *line == '\t')
                        ++line;

                      if (def)
                        {
                          if (*def != '\0')
                            {
                              if ((tmpstring = (char *) AllocVecPooled (mempool, strlen (def) + ilen + 7)))
                                {
                                  if (ilen < 9)
                                    sprintf (tmpstring, "%s\t\t= %s", item, def);
                                  else if (ilen < 18)
                                    sprintf (tmpstring, "%s\t= %s", item, def);
                                  else
                                    sprintf (tmpstring, "%s = %s", item, def);

                                  FreeVecPooled (mempool, worknode->ln_Name);
                                  worknode->ln_Name = tmpstring;

                                  return (TRUE);
                                }
                            }
                          else
                            {
                              Remove (worknode);
                              FreeVecPooled (mempool, worknode->ln_Name);
                              FreeSPooled (mempool, worknode, sizeof (struct Node));
                            }
                        }
                      else
                        {
                          Remove (worknode);
                          FreeVecPooled (mempool, worknode->ln_Name);
                          FreeSPooled (mempool, worknode, sizeof (struct Node));
                        }

                      return (FALSE);
                    }
                }
            }
        }

      worknode = worknode->ln_Succ;
    }

  /* we've reached the end of the file without finding the requested item in */
  /* the requested section, so we'll return the default value instead. */
  if (def ? *def != '\0' : FALSE)
    {
      if ((tmpstring = (char *) AllocVecPooled (mempool, (ilen = strlen (item)) + strlen (def) + 7)))
        {
          if (ilen < 9)
            sprintf (tmpstring, "%s\t\t= %s", item, def);
          else if (ilen < 18)
            sprintf (tmpstring, "%s\t= %s", item, def);
          else
            sprintf (tmpstring, "%s = %s", item, def);

          if (add_list_item (tmpstring))
            return (TRUE);
          else
            FreeVecPooled (mempool, tmpstring);
        }
    }

  return (FALSE);
}

///

/// "Get (char *item, char *def)"
char *ConfigClass::
Get (char *item, char *def)
{
  char *line, *tmp;

  if ((line = GetString (item, (char *) NULL)))
    {
      while (*line == ' ' || *line == '\t')
        line++;

      if (*line == '\"')
        {
          if ((tmp = strchr (++line, '\"')))
            {
              *tmp = '\0';
              line = newstrcpy (mempool, line);
              *tmp = '\"';
              return (line);
            }
        }
    }

  return (newstrcpy (mempool, def));
}

///

/// "Set (char *item, char *def)"

BOOL ConfigClass::
Set (char *item, char *def)
{
  char *tmpstring = NULL;
  struct Node *worknode = cfglist.lh_Head;
  int ilen;
  char *line;

  /* Now scan the file looking the desired section and item */
  while (worknode->ln_Succ != NULL)
    {
      if ((line = worknode->ln_Name))
        {
          while (*line == ' ' || *line == '\t')
            ++line;

          if (*line != '\0')    /* (strlen (worknode->ln_Name) > 0) */
            {
              if ((strncmp (line, item, ilen = strlen (item)) == 0))
                {
                  line += ilen;

                  while (*line == ' ' || *line == '\t')
                    ++line;

                  if (*line++ == '=')
                    {

                      while (*line == ' ' || *line == '\t')
                        ++line;

                      if (def)
                        {
                          if (*def != '\0')
                            {
                              if ((tmpstring = (char *) AllocVecPooled (mempool, strlen (def) + ilen + 7)))
                                {
                                  if (ilen < 9)
                                    sprintf (tmpstring, "%s\t\t= \"%s\"", item, def);
                                  else if (ilen < 18)
                                    sprintf (tmpstring, "%s\t= \"%s\"", item, def);
                                  else
                                    sprintf (tmpstring, "%s = \"%s\"", item, def);

                                  FreeVecPooled (mempool, worknode->ln_Name);
                                  worknode->ln_Name = tmpstring;

                                  return (TRUE);
                                }
                            }
                          else
                            {
                              Remove (worknode);
                              FreeVecPooled (mempool, worknode->ln_Name);
                              FreeSPooled (mempool, worknode, sizeof (struct Node));
                            }
                        }
                      else
                        {
                          Remove (worknode);
                          FreeVecPooled (mempool, worknode->ln_Name);
                          FreeSPooled (mempool, worknode, sizeof (struct Node));
                        }

                      return (FALSE);
                    }
                }
            }
        }

      worknode = worknode->ln_Succ;
    }

  /* we've reached the end of the file without finding the requested item in */
  /* the requested section, so we'll return the default value instead. */
  if (def ? *def != '\0' : FALSE)
    {
      if ((tmpstring = (char *) AllocVecPooled (mempool, (ilen = strlen (item)) + strlen (def) + 7)))
        {
          if (ilen < 9)
            sprintf (tmpstring, "%s\t\t= \"%s\"", item, def);
          else if (ilen < 18)
            sprintf (tmpstring, "%s\t= \"%s\"", item, def);
          else
            sprintf (tmpstring, "%s = \"%s\"", item, def);

          if (add_list_item (tmpstring))
            return (TRUE);
          else
            FreeVecPooled (mempool, tmpstring);
        }
    }

  return (FALSE);
}

///

/// "Get (char *item, long def)"

long ConfigClass::
Get (char *item, long def)
{
  char numbuf[20];
  char *retbuf;
  LONG temp_num;

  sprintf (numbuf, "%d", def);

  retbuf = GetString (item, numbuf);

  temp_num = atoi (retbuf);

  if (temp_num)
    return (temp_num);
  else
    return (def);

}

///

/// "Set (char *item, long def)"

BOOL ConfigClass::
Set (char *item, long def)
{
  char numbuf[20];

  sprintf (numbuf, "%d", def);

  return (SetString (item, numbuf));
}

///

/// "void Get  (char *item, char **list, long num, char **data)"
void ConfigClass::
Get (char *item, char **list, long num, char **data)
{
  char *line, *tmp;
  long i = 0;

  if ((line = GetString (item, (char *) NULL)))
    {
      for (i = 0; i < num; i++)
        {
          while (*line == ' ' || *line == '\t')
            line++;

          if (*line == '\"')
            {
              if ((tmp = strchr (++line, '\"')))
                {
                  *tmp = '\0';
                  list[i] = newstrcpy (mempool, line);
                  *tmp = '\"';
                  line = ++tmp;

                  while (*line == ' ' || *line == '\t')
                    line++;

                  if (*line == ',')
                    line++;
                }
              else
                break;
            }
          else
            break;
        }

      for (; i < num; i++)
        list[i] = NULL;
    }
  else if (data)
    {
      for (i = 0; i < num; i++)
        {
          list[i] = data[i];
        }
    }
  else
    {
      for (i = 0; i < num; i++)
        {
          list[i] = NULL;
        }
    }

}
///

/// void Set (char *item, char **list, long num)
void ConfigClass::
Set (char *item, char **list, long num)
{
  int len = 0, i;
  char *def;

  for (i = 0; i < num; i++)
    {
      len += s_strlen (list[i]);
    }

  if ((def = (char *) AllocVecPooled (mempool, len + (4 * num) + 1)))
    {
      char *end = def;
      *def = '\0';

      for (i = 0; i < num; i++)
        {
          if (list[i])
            {
              if ((i + 1 >= num) || (!list[i + 1]))
                end += sprintf (end, "\"%s\"", list[i]);
              else
                end += sprintf (end, "\"%s\", ", list[i]);
            }
        }

      SetString (item, def);

      FreeVecPooled (mempool, def);
    }
}

///

/// "void Get  (char *item, long  *list, long num, long  *data)"
void ConfigClass::
Get (char *item, long *list, long num, long *data)
{
  char *line;
  long i;

  if ((line = GetString (item, (char *) NULL)))
    {
      for (i = 0; i < num; i++)
        {
          while (*line == ' ' || *line == '\t')
            line++;

          list[i] = atol (line);

          while (isdigit (*line))
            line++;

          while (*line == ' ' || *line == '\t')
            line++;

          if (*line == ',')
            line++;
        }
    }
  else if (data)
    {
      for (i = 0; i < num; i++)
        list[i] = data[i];
    }
  else
    {
      for (i = 0; i < num; i++)
        list[i] = 0;
    }

}


///

/// void Set (char *item, long *list, long num)
void ConfigClass::
Set (char *item, long *list, long num)
{
  int len = 0;
  char *def = NULL;
  char numbuf[20];

  for (int i = 0; i < num; i++)
    {
      sprintf (numbuf, "%d", list[i]);
      len += strlen (numbuf);
    }

  if ((def = (char *) AllocVecPooled (mempool, len + (2 * num) + 1)))
    {
      char *end = def;
      *def = '\0';

      for (int i = 0; i < num; i++)
        {
          sprintf (numbuf, "%d", list[i]);

          if (i + 1 >= num)
            end += sprintf (end, "%s", numbuf);
          else
            end += sprintf (end, "%s, ", numbuf);
        }

      SetString (item, def);

      FreeVecPooled (mempool, def);
    }
}

///


/*

   List functions...

 */

/// "add_list_item (char *item)"

int ConfigClass::
add_list_item (char *item)
{
  struct Node *worknode;

  if ((worknode = (struct Node *) AllocSPooled (mempool, sizeof (struct Node))))
    {
      worknode->ln_Name = item;
      worknode->ln_Pri = 0;
      worknode->ln_Type = NT_UNKNOWN;

      AddTail (&cfglist, worknode);

      return (TRUE);
    }

  return (FALSE);
}

///
