#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <proto/dos.h>
#include <proto/exec.h>
#include <dos/dos.h>
#include <exec/memory.h>
#include <exec/lists.h>

#include "config.h"

static char buf[255];      // tymczasowe do operacji I/O

// pobiera argument z bufora, ktory znajduje sie po znaku '='
char *get_cfg_arg(char *in)
{
    char *body;

    body = strchr(in, '=');
    if(!body)
        return(NULL);
    body++;
    body = stpblk(body);
    return(strdup(body));
}

struct SimpleConfig *cfg_create_simpleconfig(char *name)
{
    struct SimpleConfig *sc;

    sc = AllocVec(sizeof(struct SimpleConfig), MEMF_ANY | MEMF_CLEAR);
    if(!sc)
        return(NULL);

    if(!(sc->sc_File = Open(name, MODE_NEWFILE)))
    {
        FreeVec(sc);
        return(NULL);
    }

    if(!(sc->sc_Entries = AllocVec(sizeof(struct List), MEMF_ANY | MEMF_CLEAR)))
    {
        Close(sc->sc_File);
        FreeVec(sc);
        return(NULL);
    }

    sc->sc_FileName = strdup(name);

    NewList(sc->sc_Entries);

    return(sc);
}

struct SimpleConfig *cfg_open_simpleconfig(char *name)
{
    struct SimpleConfig *sc;
    char token[64];

    sc = AllocVec(sizeof(struct SimpleConfig), MEMF_ANY | MEMF_CLEAR);
    if(!sc)
        return(NULL);

    if(!(sc->sc_File = Open(name, MODE_OLDFILE)))
    {
        FreeVec(sc);
        return(NULL);
    }

    if(!(sc->sc_Entries = AllocVec(sizeof(struct List), MEMF_ANY | MEMF_CLEAR)))
    {
        Close(sc->sc_File);
        FreeVec(sc);
        return(NULL);
    }

    sc->sc_FileName = strdup(name);

    NewList(sc->sc_Entries);

    while(FGets(sc->sc_File, buf, sizeof(buf)))
    {
        struct ConfigNode *cn;
        char              *arg, *ln;

        ln = strchr(buf, '\n');
        if(ln)
            *ln = 0;

        cn = AllocVec(sizeof(struct ConfigNode), MEMF_ANY | MEMF_CLEAR);

        stptok(buf, token, sizeof(token), " ");
        cn->cn_Name = strdup(token + 1);

        arg = get_cfg_arg(buf);

        if(buf[0] == 'S')
            cn->cn_SValue = strdup(arg);
        else
            cn->cn_NValue = atol(arg);

        free(arg);

        AddTail(sc->sc_Entries, cn);
    }

    return(sc);
}

void cfg_save_simpleconfig(struct SimpleConfig *sc)
{
    struct ConfigNode *cn;

    Close(sc->sc_File);
    sc->sc_File = Open(sc->sc_FileName, MODE_NEWFILE);
    if(!sc->sc_File)
        return;

    cn = (struct ConfigNode *)sc->sc_Entries->lh_Head;

    while(cn->cn_Next)
    {
        if(cn->cn_SValue)
            sprintf(buf, "S%s = %s\n", cn->cn_Name, cn->cn_SValue);
        else
            sprintf(buf, "N%s = %d\n", cn->cn_Name, cn->cn_NValue);

        FPuts(sc->sc_File, buf);

        cn = cn->cn_Next;
    }
}

int cfg_add_string_entry(struct SimpleConfig *cfg, char *name, char *value)
{
    struct ConfigNode *cn;

    cn = AllocVec(sizeof(struct ConfigNode), MEMF_ANY | MEMF_CLEAR);
    if(!cn)
        return(0);

    cn->cn_Name   = strdup(name);
    cn->cn_SValue = strdup(value);

    AddTail(cfg->sc_Entries, cn);
    return(0);
}

int cfg_add_number_entry(struct SimpleConfig *cfg, char *name, long value)
{
    struct ConfigNode *cn;

    cn = AllocVec(sizeof(struct ConfigNode), MEMF_ANY | MEMF_CLEAR);
    if(!cn)
        return(0);

    cn->cn_Name   = strdup(name);
    cn->cn_NValue = value;

    AddTail(cfg->sc_Entries, cn);
    return(0);
}

int cfg_mod_number_entry(struct SimpleConfig *cfg, char *name, long value)
{
    struct ConfigNode *cn;

    cn = (struct ConfigNode *)cfg->sc_Entries->lh_Head;

    while(cn->cn_Next)
    {
        if(!stricmp(name, cn->cn_Name))
        {
            cn->cn_NValue = value;
            break;
        }

        cn = cn->cn_Next;
    }
    return(0);
}

int cfg_mod_string_entry(struct SimpleConfig *cfg, char *name, char *value)
{
    struct ConfigNode *cn;

    cn = (struct ConfigNode *)cfg->sc_Entries->lh_Head;

    while(cn->cn_Next)
    {
        if(!stricmp(name, cn->cn_Name))
        {
            free(cn->cn_SValue),
            cn->cn_SValue = strdup(value);
            break;
        }

        cn = cn->cn_Next;
    }
    return(0);
}

char *cfg_get_string_entry(struct SimpleConfig *cfg, char *name)
{
    struct ConfigNode *cn;

    cn = (struct ConfigNode *)cfg->sc_Entries->lh_Head;

    while(cn->cn_Next)
    {
        if(!stricmp(name, cn->cn_Name))
            return(cn->cn_SValue);

        cn = cn->cn_Next;
    }
    return(0);
}

long cfg_get_number_entry(struct SimpleConfig *cfg, char *name)
{
    struct ConfigNode *cn;

    cn = (struct ConfigNode *)cfg->sc_Entries->lh_Head;

    while(cn->cn_Next)
    {
        if(!stricmp(name, cn->cn_Name))
            return(cn->cn_NValue);

        cn = cn->cn_Next;
    }
    return(0);
}

int cfg_rem_entry(struct SimpleConfig *cfg, char *name, long num)
{
    struct ConfigNode *cn;
    long               n = 0;

    cn = (struct ConfigNode *)cfg->sc_Entries->lh_Head;

    while(cn->cn_Next)
    {
        if(name)
        {
            if(!stricmp(name, cn->cn_Name))
            {
                Remove(cn);

                free(cn->cn_SValue),
                FreeVec(cn);
                break;
            }
        }
        else
        {
            if(n == num)
            {
                Remove(cn);
                FreeVec(cn);
                break;
            }
        }

        cn = cn->cn_Next;
        n++;
    }
    return(0);
}

void cfg_free_simpleconfig(struct SimpleConfig *sc)
{
    struct ConfigNode *cn, *next;

    if(!sc)
        return;

    cn = (struct ConfigNode *)sc->sc_Entries->lh_Head;

    while(cn->cn_Next)
    {
        next = cn->cn_Next;

        if(cn->cn_SValue)
            free(cn->cn_SValue);
        FreeVec(cn);

        cn = next;
    }

    if(sc->sc_File)
        Close(sc->sc_File);
    free(sc->sc_FileName);
    FreeVec(sc->sc_Entries);
    FreeVec(sc);
}

/*
int main(void)
{
    struct SimpleConfig *sc;
    struct ConfigNode   *cn;

    if(cfg_load_config() > 0)
    {
        printf("%s\n%s\n%s\n", MainConfig.cfg_User,
                            MainConfig.cfg_EMail,
                            MainConfig.cfg_NNTPHost);
        cfg_free();
    }


    sc = cfg_create_simpleconfig("ram:sc1");

    cfg_add_string_entry(sc, "Last1", "dupa maryna");
    cfg_add_number_entry(sc, "Last2", 999999);
    cfg_add_string_entry(sc, "Last3", "123");
    cfg_add_number_entry(sc, "Last4", 10000);
    cfg_add_string_entry(sc, "Last5", "ciota");

    cfg_save_simpleconfig(sc);
    cfg_free_simpleconfig(sc);

    sc = cfg_open_simpleconfig("ram:sc1");

    cn = (struct ConfigNode *)sc->sc_Entries->lh_Head;

    while(cn->cn_Next)
    {
        if(cn->cn_SValue)
            printf("S%s = %s\n", cn->cn_Name, cn->cn_SValue);
        else
            printf("N%s = %d\n", cn->cn_Name, cn->cn_NValue);

        cn = cn->cn_Next;
    }

    printf("%d\n", cfg_get_number_entry(sc, "Last4"));

    cfg_free_simpleconfig(sc);
}
*/
