/*
$Log: YAM2NN_kf.c $
Revision 1.1  1999/04/14 18:38:49  kbryd
Initial revision

*/
/*
**   YAM2NN - Usenet access for YAM p7 and newer
**   Copyright (C) 1999 Karol Bryd <kbryd@femina.com.pl>
**
**   This program is free software; you can redistribute it and/or
**   modify it under the terms of the GNU General Public License
**   as published by the Free Software Foundation; either version 2
**   of the License, or (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "Yam2NN.h"
#include "YAM2NN_kf.h"

static char rcsid[] ="$Id: YAM2NN_kf.c 1.1 1999/04/14 18:38:49 kbryd Exp kbryd $";

static char one_line[1024];

struct List *killlist;

extern char *GetEMail(char *tmp);
extern char *stristr(char *buf, char *str);

void getscore(BPTR kf, char *ngname)
{
    struct Score *score;
    char *first, *next_token;
    char token[255];
    int in_group = FALSE;

    // skip empty lines before score definitions
    while(FGets(kf, one_line, sizeof(one_line)))
    {
        if(!isspace(one_line[0]))
        {
            Seek(kf, -strlen(one_line), OFFSET_CURRENT);
            break;
        }
    }
    score = AllocVec(sizeof(struct Score), MEMF_ANY | MEMF_CLEAR);
    if(!score)
        return;
    score->newsgroup = strdup(ngname);

    while(FGets(kf, one_line, sizeof(one_line)))
    {
        first = stpblk(one_line);
        if(*first == '#')
            continue;

        if((*first == '\0') && in_group)
        {
            AddTail(killlist, (struct Node *)score);
            score = AllocVec(sizeof(struct Score), MEMF_ANY | MEMF_CLEAR);
            if(!score)
                break;
            score->newsgroup = strdup(ngname);
            in_group = FALSE;
        }
        // next score definition begins -> we have to return
        if((*first == '['))
        {
            Seek(kf, -strlen(one_line), OFFSET_CURRENT);
            return;
        }

        next_token = strchr(first, '\n');
        if(next_token)
            *next_token = 0;

        if(first[0] == '~')
        {
            score->flags = NOT;
            first++;
        }
        if(!strnicmp(first, "Subject", 7) ||
            !strnicmp(first, "Message-ID", 10) ||
            !strnicmp(first, "References", 10) ||
            !strnicmp(first, "Date", 4) ||
            !strnicmp(first, "Xref", 4) ||
            !strnicmp(first, "From", 4))
        {
            next_token   = stptok(first, token, sizeof(token), ":");
            score->field = strdup(token);

            next_token++;
            next_token = stpblk(next_token);

            next_token   = stptok(next_token, token, sizeof(token), "\r\n");
            score->sarg  = strdup(token);

            in_group = TRUE;
        }
        else
        {
            if(!(strnicmp(first, "Lines", 5)))
            {
                score->field = strdup("Lines");
                next_token = strchr(first, ':');
                if(!next_token)
                    continue;
                next_token++;
                next_token = stpblk(next_token);
                stptok(next_token, token, sizeof(token), "\r\n");
                score->narg  = atol(token);
                in_group = TRUE;
            }

            if(!(strnicmp(first, "Score", 5)))
            {
                next_token = strchr(first, ':');
                if(!next_token)
                    continue;
                next_token++;
                next_token = stpblk(next_token);
                stptok(next_token, token, sizeof(token), "\r\n");
                score->score  = atol(token);
                in_group = TRUE;
            }

            if(!(strnicmp(first, "Expires", 7)))
            {
                next_token = strchr(first, ':');
                if(!next_token)
                    continue;
                next_token++;
                next_token = stpblk(next_token);
                stptok(next_token, token, sizeof(token), "\r\n");
                sscanf(token, "%d/%d/%d", &score->expire.year,
                                          &score->expire.month,
                                          &score->expire.day);
                in_group = TRUE;
            }
        }
    }
}

void CreateKillList(void)
{
    BPTR kf;

    killlist = AllocVec(sizeof(struct List), MEMF_ANY | MEMF_CLEAR);
    if(!killlist)
        return;

    NewList(killlist);

    kf = Open("yam:killfile.scr", MODE_OLDFILE);
    if(!kf)
        return;

    while(FGets(kf, one_line, sizeof(one_line)))
    {
        static char *first, *token;

        first = stpblk(one_line);
        if((first[0] == '#') || (first[0] == '\0'))
            continue;

        // score definition begins

        if(first[0] == '[')
        {
            token = strdup(strtok(first + 1, "]"));
            getscore(kf, token);
            free(token);
        }
    }
    Close(kf);
}

void FreeKillList(void)
{
    struct Score *node, *next;

    if(!killlist)
        return;

    node = (struct Score *)killlist->lh_Head;

    if(!node)
        return;
    while(node->next)
    {
        next = node->next;

        if(node->newsgroup)
            free(node->newsgroup);
        if(node->field)
            free(node->field);
        if(node->sarg)
            free(node->sarg);

        FreeVec(node);
        node = next;
    }
}

char *GetName(char *from)
{
    char *p;

    if(p = strchr(from, '('))
        return(strtok(p + 1, ")"));
    else
    {
        p = strchr(from, '\"');
        if(p)
            return(strtok(p + 1, "\""));
        else
            return(strtok(from, "<"));
    }
}

char *ExtractSubject(char *from)
{
    static char *tmp;

    tmp = from;
    while(strncmp(tmp, "Re", 2) == 0)
    {
        if(strncmp(tmp, "Re[", 3) == 0)
        {
            tmp += 3;
            tmp = strchr(tmp, ']');
            tmp += 2;
        }
        else
            tmp += 4;
        tmp = stpblk(tmp);
    }

    tmp = stpblk(tmp);
    return(tmp);
}

char *months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
                  "Aug", "Sep", "Oct", "Nov", "Dec", 0};

struct MyDate *ConvDate(char *date_str)
{
    static struct MyDate md;
    char time[12], date[20];
    char *token;
    char tokens = 0;

    token = date_str;

    if(token[3] == ',')
    {
        token += 4;
        token = stpblk(token);
    }

    strcpy(date, "");
    strcpy(time, "");
    token = strtok(token, " ");
    while(token)
    {
        tokens++;

        if(tokens < 4)
        {
            if(tokens == 3)
            {
                if(strlen(token) == 4)
                {
                    strcat(date, token); // we want a full year, and
                    continue;            // not only the 2 numbers...:-)
                }
            }
            strcat(date, token);
            strcat(date, "-");
        }
        else
        {
            token = strtok(NULL, " ");
            if(!token)
                break;
            strcat(time, token);
            break;
        }
        token = strtok(NULL, " ");
    }

    sscanf(date, "%d-%3s-%d", &md.d, time, &md.y);

    for(md.m = 1; md.m < 13; md.m++)
        if(strnicmp(months[md.m], time, 3) == 0)
            break;
    return(&md);
}

struct MyDate *GetCurrentDate(void)
{
    static struct MyDate md;
    struct DateTime dt;
    struct DateStamp ds;
    char strdate[32];

    DateStamp(&ds);

    bzero(&dt, sizeof(struct DateTime));
    dt.dat_Stamp   = ds;
    dt.dat_Format  = FORMAT_CDN;
    dt.dat_StrDate = strdate;

    DateToStr(&dt);

    sscanf(strdate, "%d-%d-%d", &md.d, &md.m, &md.y);

    if(ds.ds_Days >= 8035) // first day of Y2K
        md.y += 2000;
    else
        md.y += 1900;

    return(&md);
}

int Compare(char *s, char *d, int val)
{
    if(val)
    {
        if(stristr(s, d))
            return(TRUE);
        if(astcsma(s, d))
            return(TRUE);
    }
    else
    {
        if(!stristr(s, d))
            return(TRUE);
        if(astcsma(s, d))
            return(TRUE);
    }
    return(FALSE);
}

int CheckArticle(struct Article *art, struct Score *score)
{
    static char local[255];
    char *s, *p;
    int val = TRUE;

    if(score->flags == NOT)
        val = FALSE;

    if(!(strnicmp(score->field, "From", 4)))
    {
        s = GetName(art->From);
        p = GetEMail(art->From);

        if(val)
        {
            if(stristr(s, score->sarg))
                return(TRUE);
            if(astcsma(s, score->sarg))
                return(TRUE);

            if(stristr(p, score->sarg))
                return(TRUE);
            if(astcsma(p, score->sarg))
                return(TRUE);
        }
        else
        {
            if(!stristr(s, score->sarg))
                return(TRUE);
            if(astcsma(s, score->sarg))
                return(TRUE);

            if(!stristr(p, score->sarg))
                return(TRUE);
            if(astcsma(p, score->sarg))
                return(TRUE);
        }
    }
    else if(!(strnicmp(score->field, "Subject", 7)))
    {
        strcpy(local, art->Subject);
        s = ExtractSubject(local);

        return(Compare(s, score->sarg, val));
    }
    else if(!(strnicmp(score->field, "MessageID", 8)))
        return(Compare(art->Msg_ID, score->sarg, val));
    else if(!(strnicmp(score->field, "References", 9)))
        return(Compare(art->References, score->sarg, val));
    else if(!(strnicmp(score->field, "Xref", 4)))
        return(Compare(art->Xref, score->sarg, val));
    else if(!(strnicmp(score->field, "Lines", 5)))
    {
        if(art->Lines >= score->narg)
            return(TRUE);
    }
    else if(!(strnicmp(score->field, "Size", 4)))
    {
        if(art->Size >= score->narg)
            return(TRUE);
    }
    else if(!(strnicmp(score->field, "Date", 4)))
    {
        struct MyDate *md;
        struct MyDate *current;

        md = ConvDate(score->sarg); // date extracted from XOVER structure
        current = GetCurrentDate(); // current date

        if(val)
        {
            if(md->y < current->y)
                return(TRUE);
            if(md->m < current->m)
                return(TRUE);
            if(md->d < current->d)
                return(TRUE);
            return(FALSE);
        }
        else
        {
            if(md->y > current->y)
                return(TRUE);
            if(md->m > current->m)
                return(TRUE);
            if(md->d > current->d)
                return(TRUE);
            return(FALSE);
        }
    }

    return(0);
}

int ScoreMessage(struct NewsNode *nn, struct Article *art)
{
    static struct Score *node;

    node = (struct Score *)killlist->lh_Head;

    while(node->next)
    {
        if(node)
        {
            if(astcsma(nn->ngname, node->newsgroup))
            {
                if(CheckArticle(art, node))
                    art->Score += node->score;
            }
        }
        node = node->next;
    }

    if(art->Score <= -9999)
        return(TRUE);
    else
        return(FALSE);
}


