/*
**   YAM2NN - Usenet access for YAM p7 and newer
**   Copyright (C) 1999 Karol Bryd <kbryd@femina.com.pl>
**
**   This program is free software; you can redistribute it and/or
**   modify it under the terms of the GNU General Public License
**   as published by the Free Software Foundation; either version 2
**   of the License, or (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "YAM2NNConfig.h"
#include "killfile.h"


/* #define KILLFILE "yam:killfile.scr" */

struct List *killlist;

static char one_line[255];

#ifdef NOT
#undef NOT
#define NOT 1
#endif

extern struct Library *MUIMasterBase;
extern APTR score_list;

#ifndef __SASC
char *stpblk(char *in)
{
    while(isspace(*in))
    {
        if(!*in)
            return(NULL);
        in++;
    }
    return(in);
}
#endif

/* finds first occurance 'str' in 'buf' - case insensitive */

char *stristr(char *buf, char *str)
{
    register int len = strlen(buf);
    register int len2 = strlen(str);
    register int a;

    for(a = 0; a <= len - len2; a++)
        if(strnicmp(&buf[a], str, len2) == 0)
            return(&buf[a]);
    return(0);
}

/* reads from 'kf' score definitions for newsgroup 'ngname' */

void getscore(BPTR kf, char *ngname)
{
    struct Score  score;
    char *first, *next_token;
    char *token;
    int in_group = FALSE;

    /* skip empty lines before score definition */
    while(FGets(kf, one_line, sizeof(one_line)))
    {
        if(!isspace(one_line[0]))
        {
            Seek(kf, -strlen(one_line), OFFSET_CURRENT);
            break;
        }
    }

    bzero((char *)(&score), (int)(sizeof(struct Score)));
    score.newsgroup = strdup(ngname);

    while(FGets(kf, one_line, sizeof(one_line)))
    {
        first = stpblk(one_line);
        if(*first == '#')
            continue;

        if((*first == '\0') && in_group)
        {
            /* we came to end of score definition, add it to the list */
            DoMethod(score_list, MUIM_NList_InsertSingle, &score, MUIV_NList_Insert_Bottom);
            score.newsgroup = strdup(ngname);
            in_group = FALSE;
        }
        /* next score definition begins -> we have to return */
        if((*first == '['))
        {
            Seek(kf, -strlen(one_line), OFFSET_CURRENT);
            return;
        }

        next_token = strchr(first, '\n');
        if(next_token)
            *next_token = 0;

        /* if field name begins with '~' then it is a 'reversed' */
        /* filter */
        if(first[0] == '~')
        {
            score.flags = NOT;
            first++;
        }

        if(!strnicmp(first, "Subject", 7) ||
            !strnicmp(first, "Message-ID", 10) ||
            !strnicmp(first, "References", 10) ||
            !strnicmp(first, "Date", 4) ||
            !strnicmp(first, "Xref", 4) ||
            !strnicmp(first, "From", 4))
        {

            next_token = strchr(first, ':');
            next_token++;
            next_token = stpblk(next_token);

            token = strtok(next_token, "\r\n");
/*            next_token   = stptok(next_token, token, sizeof(token), "\r\n"); */
            score.sarg  = strdup(token);

            token = strtok(first, ":");
/*            next_token   = stptok(first, token, sizeof(token), ":"); */
            score.field = strdup(token);

            in_group = TRUE;
        }
        else
        {
            if(!(strnicmp(first, "Lines", 5)))
            {
                score.field = strdup("Lines");
                next_token = strchr(first, ':');
                if(!next_token)
                    continue;
                next_token++;
                next_token = stpblk(next_token);
                token = strtok(next_token, "\r\n");
/*                stptok(next_token, token, sizeof(token), "\r\n"); */
                score.narg  = atol(token);
                in_group = TRUE;
            }

            if(!(strnicmp(first, "Score", 5)))
            {
                next_token = strchr(first, ':');
                if(!next_token)
                    continue;
                next_token++;
                next_token = stpblk(next_token);
                token = strtok(next_token, "\r\n");
/*                stptok(next_token, token, sizeof(token), "\r\n"); */
                score.score  = atol(token);
                in_group = TRUE;
            }

            if(!(strnicmp(first, "Expires", 7)))
            {
                next_token = strchr(first, ':');
                if(!next_token)
                    continue;
                next_token++;
                next_token = stpblk(next_token);
                token = strtok(next_token, "\r\n");
/*                stptok(next_token, token, sizeof(token), "\r\n"); */
                sscanf(token, "%d/%d/%d", &score.expire.year,
                                          &score.expire.month,
                                          &score.expire.day);
                in_group = TRUE;
            }
        }
    }
}

/* finds string 'header' in string 'in' and copies rest of line after
** 'header' to 'out'
** useful for extracting field ('Subject', 'From' etc.) values
** from messages' headers
*/

int FindOldHeader(char *in, char *out, char *header)
{
    int n = 0, i = 0;
    char *h;

    h = stristr(in, header);

    if(h)
    {
        while(h[n] != ':')
            n++;

        n += 2;

        while(1)
        {
            while(h[n] != 10)
                out[i++] = h[n++];

            if(h[n - 1] != ',')
                break;
            else
            {
                while(isspace(h[n++])) ;
                n--;
            }
        }

        out[i] = 0;
        return(TRUE);
    }
    return(FALSE);
}

/* creates kill list */
void CreateKillList(void)
{
    BPTR kf;

    /* open file with it */

    kf = Open("yam:killfile.scr", MODE_OLDFILE);
    if(!kf)
        return;

    while(FGets(kf, one_line, sizeof(one_line)))
    {
        static char *first, *token;

        /* skip any comments */
        first = stpblk(one_line);
        if((first[0] == '#') || (first[0] == '\0'))
            continue;

        /* score definition begins.
        ** Here is the format of the killfile:
        ** [newsgroup name]
        ** Score: <number>
        ** one of fields declared in 'field_names[]' in YAM2NNConfig.c
        ** optionaly there can be a field 'Expires', that definies
        ** when to expire the filter
        */

        if(first[0] == '[')
        {
            token = strdup(strtok(first + 1, "]"));
            getscore(kf, token);
            free(token);
        }
    }
    Close(kf);
}

void FreeKillList(void)
{
    DoMethod(score_list, MUIM_NList_Clear);
}

/* gets a current date and converts it to MyDate structure */

struct MyDate *GetCurrentDate(void)
{
    static struct MyDate md;
    struct DateTime dt;
    struct DateStamp ds;
    char strdate[32];

    DateStamp(&ds);

    bzero(&dt, sizeof(struct DateTime));
    dt.dat_Stamp   = ds;
    dt.dat_Format  = FORMAT_CDN;
    dt.dat_StrDate = strdate;

    DateToStr(&dt);

    sscanf(strdate, "%d-%d-%d", &md.d, &md.m, &md.y);

    if(ds.ds_Days >= 8035) /* first day of Y2K */
        md.y += 2000;
    else
        md.y += 1900;

    return(&md);
}

/* saves a new - updated - killist to a file */

void UpdateKillList(void)
{
    struct Score  *score;
    struct MyDate *current;
    struct MyDate *md;
    BPTR fh;
    int a;

    fh = Open("yam:killfile.scr", MODE_NEWFILE);

    current = GetCurrentDate();

    a = 0;
    while(1)
    {
        int expired = FALSE;

        DoMethod(score_list, MUIM_NList_GetEntry, a, &score);
        if(!score)
            break;

        a++;

        if(score->expire.year || score->expire.month || score->expire.day)
        {
            if(score->expire.year < current->y)
            {
                expired = TRUE;
                goto next;
            }
            if(score->expire.year == current->y)
            {
                if(score->expire.month < current->m)
                {
                    expired = TRUE;
                    goto next;
                }

                if(score->expire.month == current->m)
                {
                    if(score->expire.day < current->d)
                    {
                        expired = TRUE;
                        goto next;
                    }
                }
            }
        }

next:
        if(!expired)
        {
            FPrintf(fh, "[%s]\n"
                "Score: %ld\n",
                score->newsgroup,
                score->score);

            if(score->flags == NOT)
                FPrintf(fh, "~%s: ", score->field);
            else
                FPrintf(fh, "%s: ", score->field);

            if(score->sarg)
                FPrintf(fh, "%s\n", score->sarg);
            else
                FPrintf(fh, "%ld\n", score->narg);

            if(score->expire.year || score->expire.month || score->expire.day)
            {
                FPrintf(fh, "Expires: %ld/%ld/%ld\n",
                    score->expire.year, score->expire.month, score->expire.day);
            }
            FPutC(fh, '\n');
        }
    }

    Close(fh);
}

