/*
**   YAM2NN - Usenet access for YAM p7 and newer
**   Copyright (C) 1999 Karol Bryd <kbryd@femina.com.pl>
**
**   This program is free software; you can redistribute it and/or
**   modify it under the terms of the GNU General Public License
**   as published by the Free Software Foundation; either version 2
**   of the License, or (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "YAM2NNConfig.h"
#include "YAM2NNConfig_cat.h"

#include <netdb.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

char fbuf[1024];

extern struct Library *SocketBase;

struct Server *server;

/* taken from AmiTCP SDK */

int getSockAddr(const char *     host,
        int          port,
        struct sockaddr_in * addr)
{
  struct hostent * remote;

  if ((remote = gethostbyname(host)) != NULL)
    memcpy(&addr->sin_addr, remote->h_addr, sizeof addr->sin_addr);
  else if ((addr->sin_addr.s_addr = inet_addr(host)) == (unsigned long)-1) {
    return FALSE;
  }

  addr->sin_port = htons(port);
  addr->sin_family = AF_INET;

  return TRUE;
}

const int ONE = 1;

int inetconn (const struct sockaddr_in * addr)
{
  int sd;

  if ((sd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror ("socket");
    return -1;
  }

  setsockopt(sd, SOL_SOCKET, SO_KEEPALIVE, &ONE, sizeof ONE);

  if (connect(sd, (struct sockaddr *)addr, sizeof (struct sockaddr_in)) < 0) {
    close(sd);
    perror("connect");
    return -1;
  }

  return sd;
}


int do_connect(char * host, int port)
{
  struct sockaddr_in
    addr = { 0 }; /* implicit bzero() here when compiled w/ gcc */

  if (getSockAddr(host, port, &addr) == FALSE) {
    Printf("Cannot resolve host address\n");
    return -1;
  }

  return inetconn(&addr);
}

int ngets(char *p, int fd)
{
    int n;
    char *q;

    *p = 0;
    for(;;)
    {
        q = strchr(fbuf, '\n');
        if(q)
        {
            n = q - fbuf + 1;
            strncpy(p, fbuf, n);
            p += n;
            strcpy(fbuf, q+1);
            break;
        }
        n = strlen(fbuf);
        strcpy(p, fbuf);
        p += n;
        n = recv(fd, fbuf, 1024, 0);
        if(n < 0)
            return(-1);
        fbuf[n] = 0;
    }
    *p = 0;
    return 0;
}

int nwrite(int socket, char *s)
{
    int len = strlen(s);

    send(socket, s, len, 0);
    return(len);
}

int SAVEDS ASM ServerDisplay(register __a2 char **array GNUCREG(a2), register __a1 struct Server *srv GNUCREG(a1))
{
    if(!srv)
    {
        *array++ = (char *)msgName;
        *array++ = (char *)msgPortNumber;
        *array = 0;
    }
    else
    {
        static char port[4];

        *array++ = srv->srv_Server;
        sprintf(port, "%d", srv->srv_Port);
        *array++ = port;
        *array=NULL;
    }
    return(0);
}

APTR SAVEDS ASM ServerConstruct(register __a2 APTR pool GNUCREG(a2), register __a1 char *server GNUCREG(a1))
{
    struct Server *newserver;
    char *name;

    if(newserver = AllocPooled(pool, sizeof(struct Server)))
    {
        name = strtok(server, ":\n");
        newserver->srv_Server = AllocPooled(pool, strlen(name) + 1);
        strcpy(newserver->srv_Server, name);
        name = strtok(NULL, "\n\0 ");
        newserver->srv_Port = atoi(name);
    }
    return((APTR)newserver);
}

void SAVEDS ASM ServerDestruct(register __a2 APTR pool GNUCREG(a2), register __a1 struct Server *server GNUCREG(a1))
{
    FreePooled(pool, server, sizeof(struct Server));
}

struct Hook server_construct_hook = {{NULL, NULL}, (void *)(ServerConstruct), NULL, NULL};
struct Hook server_destruct_hook = {{NULL, NULL}, (void *)(ServerDestruct), NULL, NULL};
struct Hook server_display_hook = {{NULL, NULL}, (void *)(ServerDisplay), NULL, NULL};

SAVEDS struct Server *SetServer(void)
{
    struct Server *server = NULL, *selected = NULL;
    APTR dn_win, list, done, cancel;
    int a;

    set(app, MUIA_Application_Sleep, TRUE);

    dn_win = WindowObject,
        MUIA_Window_Title, msgDownloadNewsgroups,
        MUIA_Window_ID   , MAKE_ID('S','R','V','2'),
        WindowContents, VGroup,
            Child, HGroup,
                Child, VGroup, MUIA_Frame, MUIV_Frame_Group, MUIA_FrameTitle, (char *)msgAvailableHosts,
                    Child, list = NListviewObject,
                        MUIA_NList_Input, TRUE,
                        MUIA_NListview_NList, NListObject, InputListFrame,
                            MUIA_NList_Title, msgAvailableHosts,
                            MUIA_NList_Format, "BAR,BAR",
                            MUIA_NList_DisplayHook, &server_display_hook,
                            MUIA_NList_ConstructHook, &server_construct_hook,
                            MUIA_NList_DestructHook, &server_destruct_hook,
                            End,
                        End,
                    Child, HGroup,
                        Child, done = SimpleButton(msgOk),
                        Child, cancel = SimpleButton(msgCancel),
                    End,
                End,
            End,
        End,
    End;

    DoMethod(app, OM_ADDMEMBER, dn_win);

    for(a = 0;; a++)
    {
        char server_name[50];

        DoMethod(server_list, MUIM_NList_GetEntry, a, &server);
        if(!server)
            break;

        sprintf(server_name, "%s:%d", server->srv_Server, server->srv_Port);
        DoMethod(list, MUIM_NList_InsertSingle, server_name);
    }

    DoMethod(list, MUIM_Notify, MUIA_NList_DoubleClick, MUIV_EveryTime,
        app, 2, MUIM_Application_ReturnID, ID_DONE);

    DoMethod(done, MUIM_Notify, MUIA_Pressed, FALSE,
        app, 2, MUIM_Application_ReturnID, ID_DONE);

    DoMethod(cancel, MUIM_Notify, MUIA_Pressed, FALSE,
        app, 2, MUIM_Application_ReturnID, MUIV_Application_ReturnID_Quit);

    set(dn_win, MUIA_Window_Open, TRUE);
    {
        ULONG sigs = 0, close;

        close=FALSE;
        while(!close)
        {
            switch (DoMethod(app, MUIM_Application_Input, &sigs))
            {
                case ID_CANCEL:
                case MUIV_Application_ReturnID_Quit:
                    close = TRUE;
                    break;
                case ID_DONE:
                    {
                        DoMethod(list, MUIM_NList_GetEntry, MUIV_NList_GetEntry_Active, &selected);
                        server = AllocVec(sizeof(struct Server), MEMF_ANY | MEMF_CLEAR);
                        server->srv_Server = AllocVec(strlen(selected->srv_Server) + 1, MEMF_ANY|MEMF_CLEAR);
                        strcpy(server->srv_Server, selected->srv_Server);
                        server->srv_Port = selected->srv_Port;
                        close = TRUE;
                    }
                    break;
            }
            if (sigs)
                sigs = Wait(sigs);
        }
    }

    set(dn_win, MUIA_Window_Open, FALSE);
    DoMethod(app, OM_REMMEMBER, dn_win);
    MUI_DisposeObject(dn_win);
    set(app, MUIA_Application_Sleep, FALSE);
    return(server);
}


void SAVEDS DownLoadNewsgroups(void)
{
    struct Server primary_host;
    char *buf, *name, *feed;
    int sd, a = 0, b = 0;
    BPTR fh;

    abort_download = FALSE;

    if(!SocketBase)
        return;

    server = SetServer();
    if(!server)
        return;

    primary_host.srv_Server = AllocVec(strlen(server->srv_Server) + 1, MEMF_ANY|MEMF_CLEAR);
    strcpy(primary_host.srv_Server, server->srv_Server);
    primary_host.srv_Port   = server->srv_Port;

    FreeVec(server->srv_Server);
    FreeVec(server);

    set(transfer_win, MUIA_Window_Open, TRUE);

    sd = do_connect(primary_host.srv_Server, primary_host.srv_Port);
    if(sd < 0)
    {
        DisplayRequester((char *)msgCouldnotConnect, (char *)msgOk);
        set(transfer_win, MUIA_Window_Open, FALSE);
        return;
    }

    buf = AllocVec(1024, MEMF_ANY|MEMF_CLEAR);

    ngets(buf, sd);

    nwrite(sd, "mode reader\r\n");
    ngets(buf, sd);

    nwrite(sd, "LIST\r\n");
    ngets(buf, sd);

    set(unselected_newsgroups, MUIA_NList_Quiet, MUIV_NList_Quiet_Visual);

    DoMethod(unselected_newsgroups, MUIM_NList_Clear);

    a = 0;
    while(1)
    {
        ngets(buf, sd);
        if(strcmp(buf, ".\r\n") == 0)
            break;

        if(abort_download)
            break;

        feed = strchr(buf, ' ');
        if(feed)
        {
            *feed = 10;
            *(feed + 1) = 0;
        }

        DoMethod(unselected_newsgroups, MUIM_NList_InsertSingle, buf, MUIV_NList_Insert_Sorted);
        a++;
        if(b == 100)
        {
            DoMethod(app, MUIM_Application_Input, NULL);

            sprintf(buf, msgDownloadingGroup, a);
            set(progress, MUIA_Text_Contents, buf);
            DoMethod(busy, MUIM_Busy_Move);
            b = 0;
        }
        b++;
    }


    a = 0;
    fh = Open("YAM:newsgroups.list", MODE_NEWFILE);
    if(fh)
    {
        while(1)
        {
            DoMethod(unselected_newsgroups, MUIM_NList_GetEntry, a, &name);
            if(!name)
                break;
            a++;

            sprintf(buf, "%s\n", name);
            FWrite(fh, buf, strlen(buf), 1);
        }
        Close(fh);
    }

    set(unselected_newsgroups, MUIA_NList_Quiet, FALSE);
    set(transfer_win, MUIA_Window_Open, FALSE);

    FreeVec(buf);
    close(sd);
    CloseLibrary(SocketBase);
    SocketBase = 0;
    return;
}

