/*
**   YAM2NN - Usenet access for YAM p7 and newer
**   Copyright (C) 1999 Karol Bryd <kbryd@femina.com.pl>
**
**   This program is free software; you can redistribute it and/or
**   modify it under the terms of the GNU General Public License
**   as published by the Free Software Foundation; either version 2
**   of the License, or (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include "YAM2NNConfig.h"
#include "killfile.h"
#include "config.h"
#include "YAM2NNC_kf.h"
#include "YAM2NNConfig_cat.c"

#include "skip.h"
#include "get.h"
#include "excl.h"

long __stack = 20000;

char *pages[6], *preselect_types_array[4], *servers[10];
char server_num = 0;

char *field_names[] = {"Subject", "From", "Xref", "References", "Date", "Message-ID", "Lines", "Size", NULL};

char newsgroupfile[30], edited_ng_name[64];
char s_username[60], s_email[60], s_org[60], s_postname[60], s_hostname[60], s_max[10], s_userid[64], s_pass[64];
char s_startaction[108], s_stopaction[108], s_preselectaction[108];
char s_beforesendaction[108], s_aftersendaction[108];

char fetch_all = 0;

extern struct Library *ScoketBase;
struct Library        *MUIMasterBase = NULL;
struct ReqToolsBase   *ReqToolsBase  = NULL;
struct Library        *SocketBase    = NULL;
struct Library        *LocaleBase    = NULL;
struct IntuitionBase  *IntuitionBase    = NULL;

extern struct List    *killlist;
struct List           *DeletedList   = NULL;
void                  *listpool      = NULL;

/* application, window */

APTR app, window;

/* listviews */

APTR unselected_newsgroups, selected_newsgroups, score_list, server_list;

/* buttons */

APTR clearall, delete, cancel, save;

/* strings */

APTR maxkb, userid, pass, searchfor, poppath;
APTR username, postname, hostname, email, org, path, manual;

APTR start_action, stop_action, preselect_action;
APTR before_send_action, after_send_action;

/* check boxes */

APTR fetchall;

/* pop ups */

APTR start_action_popup, stop_action_popup, preselect_action_popup;
APTR before_send_action_popup, after_send_action_popup;

/* sliders */

APTR localsize, flushheaders, init_fetch, preselect_type_obj;

/* server managament */

APTR server_name, server_port, server_add, server_del, transfer_win;
APTR set_server, server_win, server_done, server_cancel, dn_ng, abort_dn;

/* score file */
APTR field_value, field_cycle, field_score, field_update, field_not, field_add;
APTR field_delete, score_name, score_add;

APTR busy, progress;

long flushthreshold = 0, preselect_type = 0, abort_download = FALSE;

int __inline DisplayRequester(char *body, char *but)
{
    return((int)rtEZRequestTags(body, but?but:"_Ok", NULL, NULL, RTEZ_Flags, EZREQF_CENTERTEXT, RT_Underscore, '_', TAG_END));
}

/* cleanup function */

void fail(APTR app, char *str)
{
    UpdateKillList();
    FreeKillList();

    DoMethod(selected_newsgroups, MUIM_NList_UseImage, -1, 0);

    if(app)
        MUI_DisposeObject(app);

    if(MUIMasterBase)
        CloseLibrary(MUIMasterBase);

    if(ReqToolsBase)
        CloseLibrary((struct Library *)ReqToolsBase);

    if(IntuitionBase)
        CloseLibrary((struct Library *)IntuitionBase);

    if(LocaleBase)
        CloseLibrary((struct Library *)LocaleBase);

    if(listpool)
        LibDeletePool(listpool);

    CloseLibrary(SocketBase);

    SocketBase = 0;

    if(str)
    {
        puts(str);
        exit(20);
    }
    exit(0);
}

void init(void)
{
    pages[0] = (char *)msgNewsgroups;
    pages[1] = (char *)msgFilters;
    pages[2] = (char *)msgUserinfo;
    pages[3] = (char *)msgEvents;
    pages[4] = 0;

    preselect_types_array[0] = (char *)msgAlways;
    preselect_types_array[1] = (char *)msgNever;
    preselect_types_array[2] = (char *)msgWithSkippedMessages;
    preselect_types_array[3] = 0;

    if(!(IntuitionBase = (struct IntuitionBase *)OpenLibrary("intuition.library", 36)))
        fail("Unable to open intuition.library", 0);

    if(!(LocaleBase = OpenLibrary("locale.library", 36)))
        fail("Unable to open locale.library", 0);

    if(!(ReqToolsBase = (struct ReqToolsBase *)OpenLibrary("reqtools.library", 36)))
        fail((char *)msgUnableToOpenReqtools, 0);

    if(!(MUIMasterBase = OpenLibrary(MUIMASTER_NAME, MUIMASTER_VMIN)))
        fail(NULL, (char *)msgUnableToOpenMUI);

    if(!(listpool = LibCreatePool(MEMF_ANY | MEMF_CLEAR, 4096, 4096)))
        fail((char *)msgUnableToCreatePool, 0);

    SocketBase = OpenLibrary("bsdsocket.library", 4);

    strcpy(newsgroupfile, "yam:newsgroups.list");
}

void LoadNewsgroups(void)
{
    BPTR fh;
    char buf[200];
/*    long a = 0; */

    if(!(fh = Open(newsgroupfile, MODE_OLDFILE)))
    {
        MUI_Request(app, window, 0, NULL, (char *)msgOk, (char *)msgNoNewsgroupsList);
        return;
    }
    set(unselected_newsgroups, MUIA_NList_Quiet, MUIV_NList_Quiet_Visual);

    while(FGets(fh, buf, 200))
    {
/*        a++;
**        if(a > 1000)
**            break;
*/
        buf[strlen(buf) - 1] = 0;
        DoMethod(unselected_newsgroups, MUIM_NList_InsertSingle, buf, MUIV_NList_Insert_Bottom);
    }
    set(unselected_newsgroups, MUIA_NList_Quiet, FALSE);
    Close(fh);
}

/* copies grup from 'unselected_newsgroup' list to 'selected_newsgroup' list */

void SAVEDS ASM AddNewsGroup(void)
{
    char *ng;
    long pos;

    get(unselected_newsgroups, MUIA_NList_Active, &pos);
    DoMethod(unselected_newsgroups, MUIM_NList_GetEntry, pos, &ng);
    DoMethod(selected_newsgroups, MUIM_NList_InsertSingle, ng, MUIV_NList_Insert_Bottom);
}

static struct Hook addng_hook = {{NULL, NULL}, (void *)(AddNewsGroup), NULL, NULL};

/* manual adding */

void SAVEDS ASM AddNewsGroupManual(void)
{
    char *ng, *user;
    long i;

    get(manual, MUIA_String_Contents, &user);

    /* if wildcard */
    if(strchr(user, '#') || strchr(user, '*'))
    {
        /* then add all that match that wildcard */
        for(i = 0 ;; i++)
        {
            DoMethod(unselected_newsgroups, MUIM_NList_GetEntry, i, &ng);
            if(!ng)
                break;

            if(MatchPatternNoCase(user, ng))
                DoMethod(selected_newsgroups, MUIM_NList_InsertSingle, ng, MUIV_NList_Insert_Bottom);
        }
    }
    else
        DoMethod(selected_newsgroups, MUIM_NList_InsertSingle, user, MUIV_NList_Insert_Bottom);
}

static struct Hook addmanual_hook = {{NULL, NULL}, (void *)(AddNewsGroupManual), NULL, NULL};

void SAVEDS ASM ObjStrFunc(register __a2 APTR pop GNUCREG(a2), register __a1 APTR str GNUCREG(a1))
{
    char *x;

    DoMethod(pop,MUIM_List_GetEntry,MUIV_List_GetEntry_Active,&x);
    set(str, MUIA_String_Contents,x);
}

void SAVEDS AddServer(void)
{
    char fullname[32];
    char *name, *port;

    get(server_name, MUIA_String_Contents, &name);
    get(server_port, MUIA_String_Contents, &port);

    strcpy(fullname, name);
    strcat(fullname, ":");
    strcat(fullname, port);
    DoMethod(server_list, MUIM_NList_InsertSingle, fullname, MUIV_NList_Insert_Bottom);
}

static struct Hook server_add_hook = {{NULL, NULL}, (void *)(AddServer), NULL, NULL};


int ASM SAVEDS SelectServer(register __a1 int *data GNUCREG(a1))
{
    struct Server *srv;
    struct SelectedNG *ng;

    DoMethod(server_list, MUIM_NList_GetEntry, MUIV_NList_GetEntry_Active, &srv);
    if(!srv)
        DoMethod(server_list, MUIM_NList_GetEntry, 1, &srv);

    DoMethod(selected_newsgroups, MUIM_NList_GetEntry, MUIV_NList_GetEntry_Active, &ng);
    if(!ng)
    {
        DoMethod(app, MUIM_Application_ReturnID, ID_CANCEL);
        set(server_win, MUIA_Window_Open, FALSE);
        return(0);
    }

    strcpy(ng->ng_Server, srv->srv_Server);
    ng->ng_Port = srv->srv_Port;
    DoMethod(selected_newsgroups, MUIM_NList_Redraw, MUIV_NList_Redraw_Active);
    if(*data == 100)
        DoMethod(app, MUIM_Application_ReturnID, ID_CANCEL);
}

static struct Hook select_server_hook = {{NULL, NULL}, (void *)(SelectServer), NULL, NULL};

void SAVEDS SetNGServer(void)
{
    struct SelectedNG *ng;
    char title[100], port[10], *str;
    long pos;

    get(selected_newsgroups, MUIA_NList_Active, &pos);
    if(pos != MUIV_NList_Active_Off)
    {
        DoMethod(selected_newsgroups, MUIM_NList_GetEntry, MUIV_NList_GetEntry_Active, &ng);
        sprintf(title, (char *)msgSettings, ng->ng_Name);
        set(server_win, MUIA_Window_Title, title);

        sprintf(port, "%ld", ng->ng_Port);

        set(server_port, MUIA_String_Contents, port);
        set(server_name, MUIA_String_Contents, ng->ng_Server);

        set(maxkb, MUIA_Numeric_Value, ng->ng_MaxLength);
        set(init_fetch, MUIA_Numeric_Value, ng->ng_InitFetch);
        set(localsize, MUIA_Numeric_Value, ng->ng_MaxMessages);

        set(flushheaders, MUIA_Numeric_Value, ng->ng_FlushThreshold);
        set(preselect_type_obj, MUIA_Cycle_Active, ng->ng_PreselectType);
    }
    else
    {
        sprintf(title, (char *)msgSettings, "none");
        set(server_win, MUIA_Window_Title, title);
    }

    set(server_win, MUIA_Window_Open, TRUE);

    {
        ULONG sigs = 0, close;

        close = FALSE;
        while(!close)
        {
            switch (DoMethod(app, MUIM_Application_Input, &sigs))
            {
                case ID_CANCEL:
                case MUIV_Application_ReturnID_Quit:
                    get(flushheaders, MUIA_Numeric_Value, &ng->ng_FlushThreshold);
                    get(preselect_type_obj, MUIA_Cycle_Active, &ng->ng_PreselectType);
                    get(localsize, MUIA_Numeric_Value, &ng->ng_MaxMessages);
                    get(init_fetch, MUIA_Numeric_Value, &ng->ng_InitFetch);
                    get(maxkb, MUIA_Numeric_Value, &ng->ng_MaxLength);

//                    get(server_port, MUIA_String_Contents, &str);
//                    ng->ng_Port = atol(str);

//                    get(server_name, MUIA_String_Contents, &str);
//                    strncpy(ng->ng_Server, str, 200);
                    close = TRUE;
                    break;
            }
            if (sigs)
                sigs = Wait(sigs);
        }
    }

    set(server_win, MUIA_Window_Open, FALSE);
}

static struct Hook set_server_hook = {{NULL, NULL}, (void *)(SetNGServer), NULL, NULL};

int SAVEDS ASM ChangeStatus(register __a2 char **array GNUCREG(a2), register __a1 long *pos GNUCREG(a1))
{
    struct SelectedNG *ng;

    DoMethod(selected_newsgroups, MUIM_NList_GetEntry, *pos, &ng);

    if(ng)
    {
        if(ng->ng_Status == STATUS_ACTIVE)
            ng->ng_Status = STATUS_INACTIVE;
        else
            ng->ng_Status = STATUS_ACTIVE;

        DoMethod(selected_newsgroups, MUIM_NList_Redraw, *pos);
    }
    return(0);
}

static struct Hook change_status_hook = {{NULL, NULL}, (void *)(ChangeStatus), NULL, NULL};

/*********************** newsgroup hooks *************************/

APTR SAVEDS ASM ListConstruct(register __a2 APTR pool GNUCREG(a2), register __a1 char *ng GNUCREG(a1))
{
    struct SelectedNG *newng;

    if(newng = AllocPooled(pool, sizeof(struct SelectedNG)))
    {
        newng->ng_MaxLength = 0;
        newng->ng_MaxMessages = 0;
        newng->ng_Status = STATUS_ACTIVE;

        newng->ng_Name = AllocPooled(pool, 200);
        memset(newng->ng_Name, 0, 200);

        newng->ng_Path = AllocPooled(pool, 200);
        memset(newng->ng_Path, 0, 200);

        strcpy(newng->ng_Name, ng);

        newng->ng_Server = AllocPooled(pool, 200);
        strcpy(newng->ng_Server, (char *)msgNone);
        newng->ng_InitFetch = 100;
    }
    return((APTR)newng);
}

void SAVEDS ASM ListDestruct(register __a2 APTR pool GNUCREG(a2), register __a1 struct SelectedNG *ng GNUCREG(a1))
{
    struct Node *node;

    FreePooled(pool, ng->ng_Server, 200);
    FreePooled(pool, ng, sizeof(struct SelectedNG));
}

int SAVEDS ASM ListDisplay(register __a2 char **array GNUCREG(a2), register __a1 struct SelectedNG *ng GNUCREG(a1))
{
    if(!ng)
    {
        *array++ = " ";
        *array++ = (char *)msgNewsgroupName;
        *array++ = (char *)msgHostName;
        *array = 0;
    }
    else
    {
        if(ng->ng_Status == STATUS_ACTIVE)
            *array++ = "\033o[1]";
        else
            *array++ = "\033o[0]";

        *array++ = ng->ng_Name;
        *array++ = ng->ng_Server;
        *array=NULL;
    }
    return(0);
}

static struct Hook display_hook = {{NULL, NULL}, (void *)(ListDisplay), NULL, NULL};
static struct Hook construct_hook = {{NULL, NULL}, (void *)(ListConstruct), NULL, NULL};
static struct Hook destruct_hook = {{NULL, NULL}, (void *)(ListDestruct), NULL, NULL};

/*********************** end of newsgroup hook **********************/


/**************************** filter hooks **************************/

APTR SAVEDS ASM FilterConstruct(register __a2 APTR pool GNUCREG(a2), register __a1 struct Score *score GNUCREG(a1))
{
    struct Score *newscore;

    if(newscore = AllocPooled(pool, sizeof(struct Score)))
    {
        newscore->newsgroup = strdup(score->newsgroup);
        newscore->field     = strdup(score->field);
        if(score->sarg)
            newscore->sarg  = strdup(score->sarg);
        else
            newscore->sarg  = NULL;
        newscore->narg      = score->narg;
        newscore->score     = score->score;
        newscore->flags     = score->flags;
        memcpy(&newscore->expire, &score->expire, sizeof(struct expire));
    }
    return((APTR)newscore);
}

void SAVEDS ASM FilterDestruct(register __a2 APTR pool GNUCREG(a2), register __a1 struct Score *score GNUCREG(a1))
{
    free(score->newsgroup);
    free(score->field);
    if(score->sarg)
        free(score->sarg);
    FreePooled(pool, score, sizeof(struct Score));
}

int SAVEDS ASM FilterDisplay(register __a2 char **array GNUCREG(a2), register __a1 struct Score *score GNUCREG(a1))
{
    static char tmp[20], tmp2[20];

    if(score)
    {
        *array++ = score->newsgroup;
        *array++ = score->field;

        if(score->sarg)
            *array++ = score->sarg;
        else
        {
            sprintf(tmp2, "%ld", score->narg);
            *array++ = tmp2;
        }

        sprintf(tmp, "%ld", score->score);
        *array++ = tmp;

        *array = NULL;
    }
    else
    {
        *array++ = (char *)msgNewsgroup;
        *array++ = (char *)msgField;
        *array++ = (char *)msgValue;
        *array++ = (char *)msgScore;
        *array   = NULL;
    }
    return(0);
}

static struct Hook filter_display_hook = {{NULL, NULL}, (void *)(FilterDisplay), NULL, NULL};
static struct Hook filter_construct_hook = {{NULL, NULL}, (void *)(FilterConstruct), NULL, NULL};
static struct Hook filter_destruct_hook = {{NULL, NULL}, (void *)(FilterDestruct), NULL, NULL};

/*************************** end of filter hooks *********************/

static struct Hook saveconfig_hook = {{NULL, NULL}, (void *)(SaveConfig), NULL, NULL};

/* sets a new path for a selected entry in 'selected_newsgroups' list*/

void SAVEDS ASM GetPath(void)
{
    long pos;
    struct SelectedNG *ng;

    get(selected_newsgroups, MUIA_NList_Active, &pos);
    DoMethod(selected_newsgroups, MUIM_NList_GetEntry, pos, &ng);
    set(path, MUIA_String_Contents, ng->ng_Path);
}

static struct Hook getpath_hook = {{NULL, NULL}, (void *)(GetPath), NULL, NULL};

/* puts path from a selected newsgroup to the string object */
void SAVEDS ASM SetPath(void)
{
    long pos;
    struct SelectedNG *ng;
    char *newpath;

    get(selected_newsgroups, MUIA_NList_Active, &pos);
    DoMethod(selected_newsgroups, MUIM_NList_GetEntry, pos, &ng);

    get(path, MUIA_String_Contents, &newpath);
    strcpy(ng->ng_Path, newpath);
}

static struct Hook setpath_hook = {{NULL, NULL}, (void *)(SetPath), NULL, NULL};

/* removes one newsgroup from 'selected_newsgroups' list */
void SAVEDS DeleteEntry(void)
{
    struct Node *node;
    struct SelectedNG *entry;
    long pos;

    if(!DeletedList)
    {
        DeletedList = LibAllocPooled(listpool, sizeof(struct List));
        NewList(DeletedList);
    }

    get(selected_newsgroups, MUIA_NList_Active, &pos);
    if(pos != MUIV_NList_Active_Off)
    {
        DoMethod(selected_newsgroups, MUIM_NList_GetEntry, pos, &entry);
        node = LibAllocPooled(listpool, sizeof(struct Node));
        node->ln_Name = LibAllocPooled(listpool, strlen(entry->ng_Name) + 1);
        strcpy(node->ln_Name, entry->ng_Name);
        AddTail(DeletedList, node);
        DoMethod(selected_newsgroups, MUIM_NList_Remove, pos);
    }
}

static struct Hook delete_hook = {{NULL, NULL}, (void *)(DeleteEntry), NULL, NULL};

void SAVEDS DeleteScore(void)
{
    long pos;

    get(score_list, MUIA_NList_Active, &pos);
    if(pos != MUIV_NList_Active_Off)
        DoMethod(score_list, MUIM_NList_Remove, MUIV_NList_Remove_Active);
}

static struct Hook delete_score_hook = {{NULL, NULL}, (void *)(DeleteScore), NULL, NULL};

void SAVEDS UpdateScore(void)
{
    struct Score *score;
    long          pos;
    char         *str;

    get(selected_newsgroups, MUIA_NList_Active, &pos);
    if(pos != MUIV_NList_Active_Off)
    {
        DoMethod(score_list, MUIM_NList_GetEntry, MUIV_NList_GetEntry_Active, &score);

        get(score_name, MUIA_String_Contents, &str);
        free(score->newsgroup);
        score->newsgroup = strdup(str);

        get(field_cycle, MUIA_Cycle_Active, &pos);
        free(score->field);
        score->field     = strdup(field_names[pos]);

        get(field_value, MUIA_String_Contents, &str);

        if(score->sarg)
            free(score->sarg);

        if(strlen(str) > 1)
            score->sarg  = strdup(str);

        get(field_score, MUIA_String_Contents, &str);

        score->score     = atol(str);

        get(field_not, MUIA_Selected, &pos);
        if(pos)
            score->flags = NOT;
        else
            score->flags = 0;

        DoMethod(score_list, MUIM_NList_Redraw, MUIV_NList_Redraw_Active);
    }
}

static struct Hook update_score_hook = {{NULL, NULL}, (void *)(UpdateScore), NULL, NULL};

void SAVEDS AddScore(void)
{
    static struct Score *score;
    long          pos;
    char         *str;

    score = AllocVec(sizeof(struct Score), MEMF_ANY | MEMF_CLEAR);
    if(!score)
        return;

    get(score_name, MUIA_String_Contents, &str);
    score->newsgroup = strdup(str);

    get(field_cycle, MUIA_Cycle_Active, &pos);
    score->field     = strdup(field_names[pos]);

    get(field_value, MUIA_String_Contents, &str);

    if(strlen(str) > 1)
        score->sarg  = strdup(str);

    get(field_score, MUIA_String_Contents, &str);

    score->score     = atol(str);

    get(field_not, MUIA_Selected, &pos);
    if(pos)
        score->flags = NOT;
    else
        score->flags = 0;

    DoMethod(score_list, MUIM_NList_InsertSingle, score, MUIV_NList_Insert_Bottom);
}

static struct Hook add_score_hook = {{NULL, NULL}, (void *)(AddScore), NULL, NULL};

void SAVEDS ChangeMaxKB(void)
{
    struct SelectedNG *ng;

    DoMethod(selected_newsgroups, MUIM_NList_GetEntry, MUIV_NList_GetEntry_Active, &ng);
    if(ng)
        get(maxkb, MUIA_Numeric_Value, &ng->ng_MaxLength);
}

static struct Hook change_maxkb_hook = {{NULL, NULL}, (void *)(ChangeMaxKB), NULL, NULL};

void SAVEDS ChangeInitFetch(void)
{
    struct SelectedNG *ng;

    DoMethod(selected_newsgroups, MUIM_NList_GetEntry, MUIV_NList_GetEntry_Active, &ng);
    if(ng)
        get(init_fetch, MUIA_Numeric_Value, &ng->ng_InitFetch);
}

static struct Hook change_initfetch_hook = {{NULL, NULL}, (void *)(ChangeInitFetch), NULL, NULL};

void SAVEDS ChangeLocalSize(void)
{
    struct SelectedNG *ng;

    DoMethod(selected_newsgroups, MUIM_NList_GetEntry, MUIV_NList_GetEntry_Active, &ng);
    if(ng)
        get(localsize, MUIA_Numeric_Value, &ng->ng_MaxMessages);
}

static struct Hook change_localsize_hook = {{NULL, NULL}, (void *)(ChangeLocalSize), NULL, NULL};

void SAVEDS ClearFilter(void)
{
    DoMethod(score_list, MUIM_NList_Clear);
}

static struct Hook clearfilter_hook = {{NULL, NULL}, (void *)(ClearFilter), NULL, NULL};

SAVEDS void SetString(void)
{
    struct Score *score;
    static char scr[12];
    int a;

    DoMethod(score_list, MUIM_NList_GetEntry, MUIV_NList_GetEntry_Active, &score);

    if(!score)
        return;

    for(a = 0; a < 8; a++)
        if(stricmp(field_names[a], score->field) == 0)
            break;

    if(score->flags == NOT)
        set(field_not, MUIA_Selected, TRUE);
    else
        set(field_not, MUIA_Selected, FALSE);

    set(score_name, MUIA_String_Contents, score->newsgroup);
    set(field_cycle, MUIA_Cycle_Active, a);
    sprintf(scr, "%ld", score->score);
    set(field_score, MUIA_String_Contents, scr);

    if(score->sarg)
        set(field_value, MUIA_String_Contents, score->sarg);
    else
    {
        static char tmp[10];

        sprintf(tmp, "%ld", score->narg);
        set(field_value, MUIA_String_Contents, tmp);
    }
}

static struct Hook setstring_hook = {{NULL, NULL}, (void *)(SetString), NULL, NULL};

SAVEDS void RefreshServer(long sel)
{
    struct Server *server;
    static char port[8];

    DoMethod(server_list, MUIM_NList_GetEntry, MUIV_NList_GetEntry_Active, &server);

    set(server_name, MUIA_String_Contents, server->srv_Server);
    sprintf(port, "%ld", server->srv_Port);
    set(server_port, MUIA_String_Contents, port);
}

static struct Hook refresh_server_hook = {{NULL, NULL}, (void *)(RefreshServer), NULL, NULL};


static struct Hook download_ng_hook = {{NULL, NULL}, (void *)(DownLoadNewsgroups), NULL, NULL};

/* is called when user clicks on Abort button */

SAVEDS void Abort(void)
{
    abort_download = TRUE;
}

static struct Hook abort_hook = {{NULL, NULL}, (void *)(Abort), NULL, NULL};

ASM SAVEDS ULONG AslHook(register __a1 struct TagItem *tags)
{
    tags->ti_Tag = ASLFR_InitialDrawer;
    tags->ti_Data = (ULONG)"YAM:";
    return(TRUE);
}

static struct Hook asl_hook = {{NULL, NULL}, (void *)(AslHook), NULL, NULL};

void init_app(void)
{
    app = ApplicationObject,
        MUIA_Application_Title      , "Yam2NNConfig",
        MUIA_Application_Version    , "$VER: Yam2NNConfig 1.03 (23-Mar-99)",
        MUIA_Application_Copyright  , "1998/1998, Karol Bryd",
        MUIA_Application_Author     , "Karol Bryd",
        MUIA_Application_Description, msgYam2NNConfiguration,
        MUIA_Application_Base       , "Yam2NN",

        SubWindow, window = WindowObject,
            MUIA_Window_Title, msgYam2NNConfiguration,
            MUIA_Window_ID   , MAKE_ID('A','B','O','1'),
            WindowContents, VGroup,
                Child, RegisterGroup(pages),
                    MUIA_Register_Frame, TRUE,
                    Child, HGroup,
                        Child, VGroup, MUIA_Frame, MUIV_Frame_Group, MUIA_FrameTitle, msgUnselectedNewsgroups,
                            Child, dn_ng = SimpleButton((char *)msgDownloadNewsgroups),
                            Child, unselected_newsgroups = NListviewObject,
                                        MUIA_NList_Input, TRUE,
                                        MUIA_NListview_NList, NListObject, InputListFrame,
                                            MUIA_NList_ConstructHook, MUIV_NList_ConstructHook_String,
                                            MUIA_NList_DestructHook, MUIV_NList_DestructHook_String,
                                            End,
                                        End,
                            Child, HGroup,
                                Child, Label(msgManualAdding), Child, manual = String(0, 64),
                                Child, Label(msgSearchFor), Child, searchfor = String(0, 64),
                            End,
                        End,

                        Child, VGroup, MUIA_Frame, MUIV_Frame_Group, MUIA_FrameTitle, msgSelectedNewsgroups,
                            Child, selected_newsgroups = NListviewObject,
                                    MUIA_NList_Input, TRUE,
                                    MUIA_NListview_NList, NListObject, InputListFrame,
                                            MUIA_NList_AutoVisible, TRUE,
                                            MUIA_NList_ConstructHook, &construct_hook,
                                            MUIA_NList_DisplayHook, &display_hook,
                                            MUIA_NList_DestructHook, &destruct_hook,
                                            MUIA_NList_Format, "BAR,BAR,BAR",
                                            MUIA_NList_Title, TRUE,
                                        End,
                                    End,
                                Child, HGroup,
                                    Child, Label(msgPath),
                                    Child, poppath = PopaslObject,
                                        MUIA_Popstring_String, path = KeyString("YAM:", 256, 'd'),
                                        MUIA_Popstring_Button, PopButton(MUII_PopDrawer),
                                        MUIA_Popasl_StartHook, &asl_hook,
                                        ASLFR_TitleText  , msgSelectDrawer,
                                        ASLFR_DrawersOnly, TRUE,
                                        ASLFR_DoSaveMode, TRUE,
                                    End,
                                End,
                                Child, clearall = SimpleButton(msgClearList),
                                Child, delete = SimpleButton(msgDeleteEntry),
                                Child, set_server = SimpleButton((char *)msgGroupPreferences),
                        End,
                    End,
                    Child, VGroup,
                        Child, HGroup,
                            Child, VGroup,
                                Child, score_list = NListviewObject,
                                    MUIA_NList_Input, TRUE,
                                    MUIA_NListview_NList, NListObject, InputListFrame,
                                            MUIA_NList_ConstructHook, &filter_construct_hook,
                                            MUIA_NList_DisplayHook, &filter_display_hook,
                                            MUIA_NList_DestructHook, &filter_destruct_hook,
                                            MUIA_NList_Format, "BAR,BAR,BAR,BAR",
                                            MUIA_NList_Title, TRUE,
                                           End,
                                    End,
                            End,
                        End,
                        Child, ColGroup(2), GroupFrameT((char *)msgSelectedFilter),
                            Child, Label((char *)msgNewsgroupName),
                            Child, score_name = String(0, 100),
                            Child, Label((char *)msgFieldName),
                            Child, field_cycle = Cycle(field_names),
                            Child, Label((char *)msgFieldValue),
                            Child, HGroup,
                                Child, field_value = String(0, 100),
                                Child, field_not = CheckMark(0),
                            End,
                            Child, Label((char *)msgScore),
                            Child, field_score = String(0, 100),
                            Child, HSpace(1),
                            Child, HGroup,
                                Child, field_add = SimpleButton((char *)msgAdd),
//                                Child, field_update = SimpleButton((char *)msgUpdate),
                                Child, field_delete = SimpleButton((char *)msgDelete),
                            End,
                        End,
                    End,

                    Child, VGroup,
                        Child, ColGroup(2),
                            Child, Label(msgUsername), Child, username     = StringObject, StringFrame, MUIA_String_MaxLen, sizeof(s_username), MUIA_String_Contents, s_username, End,
                            Child, Label(msgHostName), Child, hostname     = StringObject, StringFrame, MUIA_String_MaxLen, sizeof(s_hostname), MUIA_String_Contents, s_hostname, End,
                            Child, Label(msgPostHostName), Child, postname = StringObject, StringFrame, MUIA_String_MaxLen, sizeof(s_postname), MUIA_String_Contents, s_postname, End,
                            Child, Label(msgEMail), Child, email           = StringObject, StringFrame, MUIA_String_MaxLen, sizeof(s_email),    MUIA_String_Contents, s_email,    End,
                            Child, Label(msgOrganization),  Child, org     = StringObject, StringFrame, MUIA_String_MaxLen, sizeof(s_org),      MUIA_String_Contents, s_org,      End,
                            Child, Label(msgUserId), Child, userid         = StringObject, StringFrame, MUIA_String_MaxLen, 99,                 MUIA_String_Contents, s_userid,   End,
                            Child, Label(msgPassword), Child, pass         = StringObject, StringFrame, MUIA_String_MaxLen, 99,                 MUIA_String_Contents, s_pass,     MUIA_String_Secret, TRUE, End,
                        End,

                        Child, MUI_NewObject (MUIC_Group,
                            MUIA_Group_Horiz, TRUE,
                            MUIA_Group_Child, MUI_NewObject (MUIC_Rectangle, TAG_END),
                            MUIA_Group_Child, MUI_NewObject (MUIC_Text,
                            MUIA_Text_Contents, (long)"'Fetch all and save' mode",
                            MUIA_Frame, MUIV_Frame_ImageButton,
                            MUIA_FramePhantomHoriz, TRUE,
                            TAG_END),

                        Child, fetchall = MUI_NewObject (MUIC_Image,
                           MUIA_Image_Spec, MUII_CheckMark,
                           MUIA_Frame, MUIV_Frame_ImageButton,
                           MUIA_InputMode, MUIV_InputMode_Toggle,
                           MUIA_ShowSelState, FALSE,
                           MUIA_CycleChain, TRUE,
                           MUIA_Disabled, fetch_all,
                           TAG_END),

                        Child, MUI_NewObject (MUIC_Rectangle, TAG_END),
                        End,

                    End,
                    Child, VGroup,
                        Child, ColGroup(2),
                            Child, Label(msgStartAction),
                                    Child, start_action_popup = PopaslObject,
                                        MUIA_Popstring_String, start_action = String(s_startaction, sizeof(s_startaction)),
                                        MUIA_Popstring_Button, PopButton(MUII_PopFile),
                                        ASLFR_TitleText, msgSelectDrawer,
                                    End,

                            Child, Label(msgStopAction),
                                    Child, stop_action_popup = PopaslObject,
                                        MUIA_Popstring_String, stop_action = String(s_stopaction, sizeof(s_stopaction)),
                                        MUIA_Popstring_Button, PopButton(MUII_PopFile),
                                        ASLFR_TitleText, msgSelectDrawer,
                                    End,

                            Child, Label(msgBeforeSendAction),
                                    Child, before_send_action_popup = PopaslObject,
                                        MUIA_Popstring_String, before_send_action = String(s_preselectaction, sizeof(s_preselectaction)),
                                        MUIA_Popstring_Button, PopButton(MUII_PopFile),
                                        ASLFR_TitleText, msgSelectDrawer,
                                    End,
                            Child, Label(msgAfterSendAction),
                                    Child, after_send_action_popup = PopaslObject,
                                        MUIA_Popstring_String, after_send_action = String(s_preselectaction, sizeof(s_preselectaction)),
                                        MUIA_Popstring_Button, PopButton(MUII_PopFile),
                                        ASLFR_TitleText, msgSelectDrawer,
                                    End,
                            Child, Label(msgPreselectAction),
                                    Child, preselect_action_popup = PopaslObject,
                                        MUIA_Popstring_String, preselect_action = String(s_preselectaction, sizeof(s_preselectaction)),
                                        MUIA_Popstring_Button, PopButton(MUII_PopFile),
                                        ASLFR_TitleText, msgSelectDrawer,
                                    End,
                        End,
                    End,
                End,
                Child, HGroup,
                    Child, save = SimpleButton(msgSave), Child, cancel = SimpleButton(msgCancel),
                End,
            End,
            End,
            SubWindow, server_win = WindowObject,
                MUIA_Window_Title, (char *)msgServerSettings,
                MUIA_Window_ID   , MAKE_ID('S','R','V','1'),
                WindowContents, VGroup,
                        Child, VGroup, GroupFrameT((char *)msgAvailableHosts),
                            Child, server_list = NListviewObject,
                                MUIA_NList_Input, TRUE,
                                MUIA_NListview_NList, NListObject, InputListFrame,
                                    MUIA_NList_Title, (char *)msgAvailableHosts,
                                    MUIA_NList_DisplayHook, &server_display_hook,
                                    MUIA_NList_ConstructHook, &server_construct_hook,
                                    MUIA_NList_DestructHook, &server_destruct_hook,
                                    MUIA_NList_Format, "BAR,BAR",
                                    End,
                                End,
                            Child, ColGroup(2),
                                Child, Label(msgNNTPHostName), Child, server_name = String(0, 99),
                                Child, Label(msgPortNumber), Child, server_port = String(0, 99),
                            End,
                            Child, HGroup,
                                Child, server_add = SimpleButton(msgAddHost),
                                Child, server_del = SimpleButton(msgDeleteHost),
                            End,
                        End,

                        Child, VGroup, GroupFrameT((char *)msgGroupPreferences),
                            Child, ColGroup(2),
                                Child, Label(msgMaximumLength), Child, maxkb = SliderObject, MUIA_Numeric_Value, 0, MUIA_Numeric_Min, 0, MUIA_Numeric_Max, 300, MUIA_Numeric_Format, "%3ld kb", End,
                                Child, Label(msgMaxMessages), Child, localsize = SliderObject, MUIA_Numeric_Value, 0, MUIA_Numeric_Min, 0, MUIA_Numeric_Max, 300, MUIA_Numeric_Format, "%3ld", End,
                                Child, Label(msgInitialFetch), Child, init_fetch = SliderObject, MUIA_Numeric_Value, 0, MUIA_Numeric_Min, 0, MUIA_Numeric_Max, 500, MUIA_Numeric_Format, "%3ld", End,
                                Child, Label(msgDeleteHeadersAfter), Child, flushheaders = SliderObject, MUIA_Numeric_Value, flushthreshold, MUIA_Numeric_Min, 0, MUIA_Numeric_Max, 31, MUIA_Numeric_Format, msgDays, End,
                                Child, Label(msgOpenPreselectWindow), Child, preselect_type_obj = CycleObject, MUIA_Cycle_Entries, preselect_types_array, MUIA_Cycle_Active, preselect_type, MUIA_CycleChain, 1, End,

                            End,
                        End,
                        Child, HGroup,
                            Child, server_done = SimpleButton(msgDone),
                            Child, server_cancel = SimpleButton(msgCancel),
                        End,
                End,
            End,
            SubWindow, transfer_win = WindowObject,
                MUIA_Window_Title, msgDownloadingNewsgroups,
                MUIA_Window_ID   , MAKE_ID('S','R','V','6'),
                WindowContents, VGroup,
                        Child, progress = TextObject,
                               MUIA_Background, MUII_TextBack,
                               MUIA_Frame, MUIV_Frame_Text,
                               MUIA_Text_Contents, NULL,
                               MUIA_Text_SetMin, TRUE,
                        End,
                        Child, busy = MUI_NewObject("Busy.mcc", TAG_DONE),
                        Child, abort_dn = SimpleButton(msgAbort),
                End,
            End,

        End;

    if (!app)
        fail(app, (char *)msgFailedToCreateApp);

    DoMethod(window,
        MUIM_Window_SetCycleChain,
        unselected_newsgroups, manual, selected_newsgroups, poppath, path, clearall, delete, score_list,
        score_name, field_cycle, field_value, field_not, field_score, field_add, field_update,
        field_delete, maxkb, localsize, username, hostname,
        email, org, userid, pass, flushheaders, preselect_type, save, cancel,
        start_action, start_action_popup,
        stop_action, stop_action_popup,
        before_send_action, before_send_action_popup,
        after_send_action, after_send_action_popup,
        preselect_action, preselect_action_popup,
        fetchall,
        0);

    DoMethod(server_win, MUIM_Window_SetCycleChain,
        server_list, server_name, server_port, server_add, server_del, server_done,
        server_cancel, 0);

    DoMethod(set_server, MUIM_Notify, MUIA_Pressed, FALSE,
        set_server, 2, MUIM_CallHook, &set_server_hook);

    DoMethod(abort_dn, MUIM_Notify, MUIA_Pressed, FALSE,
        abort_dn, 2, MUIM_CallHook, &abort_hook);

    DoMethod(dn_ng, MUIM_Notify, MUIA_Pressed, FALSE,
        dn_ng, 2, MUIM_CallHook, &download_ng_hook);

    DoMethod(server_list, MUIM_Notify, MUIA_NList_Active, FALSE,
        server_list, 3, MUIM_CallHook, &refresh_server_hook, MUIV_TriggerValue);

    DoMethod(server_list, MUIM_Notify, MUIA_NList_DoubleClick, MUIV_EveryTime,
        server_list, 3, MUIM_CallHook, &select_server_hook, MUIV_TriggerValue);

    DoMethod(server_done, MUIM_Notify, MUIA_Pressed, FALSE,
        server_done, 3, MUIM_CallHook, &select_server_hook, 100);

    DoMethod(server_del, MUIM_Notify, MUIA_Pressed, FALSE,
        server_list, 2, MUIM_NList_Remove, MUIV_NList_Remove_Active);

    DoMethod(server_add, MUIM_Notify, MUIA_Pressed, FALSE,
        server_add, 2, MUIM_CallHook, &server_add_hook);

    DoMethod(window,MUIM_Notify,MUIA_Window_CloseRequest,TRUE,
        app,2,MUIM_Application_ReturnID,MUIV_Application_ReturnID_Quit);

    DoMethod(server_win,MUIM_Notify,MUIA_Window_CloseRequest,TRUE,
        app,2,MUIM_Application_ReturnID,MUIV_Application_ReturnID_Quit);

    DoMethod(server_cancel, MUIM_Notify, MUIA_Pressed, FALSE,
        app, 2, MUIM_Application_ReturnID, MUIV_Application_ReturnID_Quit);

    DoMethod(unselected_newsgroups, MUIM_Notify, MUIA_NList_DoubleClick, MUIV_EveryTime,
        unselected_newsgroups, 3, MUIM_CallHook, &addng_hook, MUIV_TriggerValue);

    DoMethod(selected_newsgroups, MUIM_Notify, MUIA_NList_DoubleClick, MUIV_EveryTime,
        selected_newsgroups, 3, MUIM_CallHook, &change_status_hook, MUIV_TriggerValue);

    DoMethod(selected_newsgroups, MUIM_Notify, MUIA_NList_Active, MUIV_EveryTime,
        score_list, 3, MUIM_Set, MUIA_NList_Active, MUIV_TriggerValue);

    DoMethod(selected_newsgroups, MUIM_Notify, MUIA_NList_Active, MUIV_EveryTime,
        selected_newsgroups, 3, MUIM_CallHook, &getpath_hook, MUIV_TriggerValue);

    DoMethod(path, MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
        path, 3, MUIM_CallHook, &setpath_hook, MUIV_TriggerValue);

    DoMethod(manual, MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
        manual, 3, MUIM_CallHook, &addmanual_hook, MUIV_TriggerValue);

    DoMethod(field_update, MUIM_Notify, MUIA_Pressed, FALSE,
        field_update, 3, MUIM_CallHook, &update_score_hook, MUIV_TriggerValue);

    DoMethod(field_add, MUIM_Notify, MUIA_Pressed, FALSE,
        field_add, 3, MUIM_CallHook, &add_score_hook, MUIV_TriggerValue);

    DoMethod(field_delete, MUIM_Notify, MUIA_Pressed, FALSE,
        field_delete, 3, MUIM_CallHook, &delete_score_hook, MUIV_TriggerValue);

    DoMethod(score_list, MUIM_Notify, MUIA_NList_Active, MUIV_EveryTime,
        score_list, 2, MUIM_CallHook, &setstring_hook);

    DoMethod(clearall, MUIM_Notify, MUIA_Pressed, FALSE,
        selected_newsgroups, 1, MUIM_NList_Clear);

    DoMethod(clearall, MUIM_Notify, MUIA_Pressed, FALSE,
        score_list, 1, MUIM_NList_Clear);

    DoMethod(delete, MUIM_Notify, MUIA_Pressed, FALSE,
        selected_newsgroups, 2, MUIM_CallHook, &delete_hook);

    DoMethod(cancel, MUIM_Notify, MUIA_Pressed, FALSE,
        app, 2, MUIM_Application_ReturnID, MUIV_Application_ReturnID_Quit);

    DoMethod(save, MUIM_Notify, MUIA_Pressed, FALSE,
        save, 2, MUIM_CallHook, &saveconfig_hook);

    DoMethod(maxkb, MUIM_Notify, MUIA_Slider_Level, MUIV_EveryTime,
        maxkb, 2, MUIM_CallHook, &change_maxkb_hook);

    DoMethod(init_fetch, MUIM_Notify, MUIA_Slider_Level, MUIV_EveryTime,
        init_fetch, 2, MUIM_CallHook, &change_initfetch_hook);

    DoMethod(localsize, MUIM_Notify, MUIA_Slider_Level, MUIV_EveryTime,
        localsize, 2, MUIM_CallHook, &change_localsize_hook);

    DoMethod(selected_newsgroups, MUIM_NList_UseImage, IMG, 0, ~0L);
    DoMethod(selected_newsgroups, MUIM_NList_UseImage, IMG2, 1, ~0L);
    DoMethod(selected_newsgroups, MUIM_NList_UseImage, IMG3, 2, ~0L);
}

int main(int argc,char *argv[])
{
    /* open libraries, init arrays, etc. */
    init();
    /* create MUI application */
    init_app();

    /* load newsgroups from a file */
    LoadNewsgroups();
    /* load killfile */
    CreateKillList();
    /* load configuration */
    getconfig();

    set(window,MUIA_Window_Open,TRUE);
    {
        ULONG sigs = 0, close;

        close = FALSE;
        while(!close)
        {
            switch (DoMethod(app,MUIM_Application_Input,&sigs))
            {
                case MUIV_Application_ReturnID_Quit:
                    close=TRUE;
                    break;
            }
            if (sigs)
                sigs = Wait(sigs);
        }
    }

    set(window,MUIA_Window_Open,FALSE);

    fail(app, NULL);
}
